/*==== T a p T e m p o L f o _ v 2 . c ================================================================================
    
    Copyright (c) 2013-2014 Harald Sabro. For non-commercial use only. Do not post other than at
    www.sabrotone.com.
    
    Project     : Tap-Tempo LFO using the ATtiny85, version 2.1
    Author      : Harald Sabro
    
    Note        : Debounce code based on article at: http://www.ganssle.com/debouncing-pt2.htm
                  
                  PWM DDS implementation inspired by:
                  http://interface.khm.de/index.php/lab/experiments/arduino-dds-sinewave-generator/
                  
                  Syncing on first tap, and selecting options via voltage divider/pot; ideas (not
                  code) borrowed from: http://www.electricdruid.net/index.php?page=projects.taplfo
                  
    Fuse bits   : CKDIV8 disabled.
                  RSTDISBL enabled.
    
    Important   : Because the reset pin is used for I/O instead of as an external reset, ISP is not
                  possible. HVSP is required to program correctly.

    Revisions   : 2.1 - Realigned the sine wave table so that it's phase matches those of the other
                        waveforms.
    
 ====================================================================================================================*/ 

/*====== Includes =====================================================================================================
=====================================================================================================================*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/atomic.h>

/*====== Defines ======================================================================================================
=====================================================================================================================*/

///////////////////////////////////////////////////////////////////////////////////////////////////
// Couldn't find where the bool_t was defined so here it is.
///////////////////////////////////////////////////////////////////////////////////////////////////

#define bool_t      uint8_t
#define false       0
#define true        1

///////////////////////////////////////////////////////////////////////////////////////////////////
// Defines related to timer setup.
///////////////////////////////////////////////////////////////////////////////////////////////////

#define CLOCK_FREQUENCY             8000000UL   // Fuse bits set to internal 8MHz clock (default), and no clock division (not default).

#define TIMER1_PRE_SCALER           32.0
#define TIMER1_FREQUENCY            0.001       // Timer1 trigger rate in Hertz.
#define TIMER1_SAMPLE_RATE          1           // Timer1 trigger rate in milliseconds.

///////////////////////////////////////////////////////////////////////////////////////////////////
// Defines related to the key debounce logic.
///////////////////////////////////////////////////////////////////////////////////////////////////

#define SWITCH_STATE_CLOSED         true
#define SWITCH_STATE_OPEN           false

#define DEBOUNCE_PRESS_TIME         10          // Millisecond count before considering the switch ON.
#define DEBOUNCE_RELEASE_TIME       20          // Millisecond count before considering the switch OFF again.

///////////////////////////////////////////////////////////////////////////////////////////////////
// Sine wave table definition. Got this from a calculator here:
// http://www.daycounter.com/Calculators/Sine-Generator-Calculator.phtml
//
// Note: The basic sine wave has been realigned so that we're starting with the lowest peak rather
//       than the mid-point. This matches the phase of the other waveforms.
///////////////////////////////////////////////////////////////////////////////////////////////////

#define WAVE_TABLE_RESOLUTION       256

uint8_t g_sine_table[WAVE_TABLE_RESOLUTION] =
{
      0,   0,   0,   0,   1,   1,   1,   2,   2,   3,   4,   5,   5,   6,   7,   9,
     10,  11,  12,  14,  15,  17,  18,  20,  21,  23,  25,  27,  29,  31,  33,  35,
     37,  40,  42,  44,  47,  49,  52,  54,  57,  59,  62,  65,  67,  70,  73,  76,
     79,  82,  85,  88,  90,  93,  97, 100, 103, 106, 109, 112, 115, 118, 121, 124,
    128, 131, 134, 137, 140, 143, 146, 149, 152, 155, 158, 162, 165, 167, 170, 173,
    176, 179, 182, 185, 188, 190, 193, 196, 198, 201, 203, 206, 208, 211, 213, 215,
    218, 220, 222, 224, 226, 228, 230, 232, 234, 235, 237, 238, 240, 241, 243, 244,
    245, 246, 248, 249, 250, 250, 251, 252, 253, 253, 254, 254, 254, 255, 255, 255,
    255, 255, 255, 255, 254, 254, 254, 253, 253, 252, 251, 250, 250, 249, 248, 246,
    245, 244, 243, 241, 240, 238, 237, 235, 234, 232, 230, 228, 226, 224, 222, 220,
    218, 215, 213, 211, 208, 206, 203, 201, 198, 196, 193, 190, 188, 185, 182, 179,
    176, 173, 170, 167, 165, 162, 158, 155, 152, 149, 146, 143, 140, 137, 134, 131,
    128, 124, 121, 118, 115, 112, 109, 106, 103, 100,  97,  93,  90,  88,  85,  82,
     79,  76,  73,  70,  67,  65,  62,  59,  57,  54,  52,  49,  47,  44,  42,  40,
     37,  35,  33,  31,  29,  27,  25,  23,  21,  20,  18,  17,  15,  14,  12,  11,
     10,   9,   7,   6,   5,   5,   4,   3,   2,   2,   1,   1,   1,   0,   0,   0
};

///////////////////////////////////////////////////////////////////////////////////////////////////
// List of possible wave forms.
///////////////////////////////////////////////////////////////////////////////////////////////////

typedef enum
{
    SineWave = 0,
    RampUpWave,
    RampDownWave,
    TriangleWave,
    SquareWave
} WAVEFORM_CHOICE;

#define TIMER0_SAMPLE_RATE          (CLOCK_FREQUENCY / WAVE_TABLE_RESOLUTION)   // 8Mhz / 256 = 31.25kHz

#define LFO_DEFAULT_RATE            1000        // In milliseconds = 1Hz

///////////////////////////////////////////////////////////////////////////////////////////////////
// Minimum and maximum range for tap tempo, not counting in any multiplier.
//
// Note: With an upper tap-tempo frequency of 20Hz and a 16th-multiplier we'll get
//       20Hz * (16 / 4) = 80Hz. That coincides nicely with a low pass filter of 2k2 / 1uF ~= 72Hz
//       to get rid of any PWM high-frequency noise.
///////////////////////////////////////////////////////////////////////////////////////////////////

#define LFO_MIN_TAP_TEMPO_RATE      10000       // In milliseconds = 0.1Hz / 10 seconds
#define LFO_MAX_TAP_TEMPO_RATE      50          // In milliseconds = 20Hz / 5/100 of a second
#define LFO_MAX_TOTAL_FREQUENCY     80          // In Hz (including max. multiplier)

///////////////////////////////////////////////////////////////////////////////////////////////////
// Defines related to clock/synchronization input.
///////////////////////////////////////////////////////////////////////////////////////////////////

#define SYNC_LOW_STATE_THRESHOLD        0x7f
#define SYNC_HIGH_STATE_THRESHOLD       0x80

///////////////////////////////////////////////////////////////////////////////////////////////////
// List of possible tap tempo multipliers.
///////////////////////////////////////////////////////////////////////////////////////////////////

typedef enum
{
    Whole = 1,
    Half = 2,
    Quarter = 4,        // 1:1 - Base tap tempo
    Eighth = 8,
    Sixteenth = 16
} TEMPO_MULTIPLIER;

///////////////////////////////////////////////////////////////////////////////////////////////////
// Defines making it easier to change pin configurations.
///////////////////////////////////////////////////////////////////////////////////////////////////

#define DATA_REG                PORTB
#define DATA_DIR_REG            DDRB
#define INPUT_PINS              PINB

#define LFO_PIN                 PB0     // PWM output from timer0 (OC0A = PB0).
#define LED_PIN                 PB1     // Outputs tempo pulses. Use for LED indication.
#define SWITCH_PIN              PB2     // Tap-tempo input (switch or relayed signal from other MCU).
//#define WAVEFORM_PIN            PB3     // ADC3 input. Controls waveform selection.
//#define MULTIPLIER_PIN          PB4     // ADC2 input. Controls tempo multiplier selection.
#define SYNC_PIN                PB5     // Pin-change interrupt. Reads clock/synchronization input pulses.

/*====== Local function prototypes ====================================================================================
=====================================================================================================================*/

bool_t GetSwitchState();

void DebounceSwitch(bool_t* pswitch_change,
                    bool_t* pswitch_state);

void SetLfoBaseFrequency(uint16_t milliseconds);

void RecalculateLfoFrequency();

void ResetBaseTempo();// __attribute__((always_inline));

void AlignWaveform();// __attribute__((always_inline));

void SetWaveform(uint8_t voltage_reading);

void SetMultiplier(uint8_t voltage_reading);

void PlotWaveform();// __attribute__((always_inline));

void StartTempoCount();

void StopTempoCount();

void TempoCountTimeout();

/*====== Global variables =============================================================================================
=====================================================================================================================*/

bool_t g_debounced_switch_state = SWITCH_STATE_OPEN;

uint32_t g_base_phase_accumulator = 0;
uint32_t g_base_duty_cycle = 0;
uint8_t g_base_table_index = 0xff;

uint32_t g_phase_accumulator = 0;
uint32_t g_duty_cycle = 0;
uint8_t g_table_index = 0;

bool_t g_is_counting_tempo = false;
uint16_t g_tempo_count = 0;
uint16_t g_base_tempo = 0;

uint8_t g_waveform_offset_count = 1;
uint8_t g_waveform_offset_index = 0;
uint32_t g_waveform_offset_table[2][4] =
{
    {0x00000000, 0x7f000000, 0x00000000, 0x00000000},   // Waveform offsets when at 1/2 multiplier.
    {0x00000000, 0x3f000000, 0x7f000000, 0xbf000000}    // Waveform offsets when at 1/4 multiplier.
};

WAVEFORM_CHOICE g_waveform = SineWave;
TEMPO_MULTIPLIER g_multiplier = Quarter;

/*====== Functions ====================================================================================================
=====================================================================================================================*/

/*====== m a i n ( ) ==================================================================================================

    Purpose     : Entry point and main loop.
    Arguments   : 
    Returns     : 
    Author      : Harald Sabro
    
=====================================================================================================================*/

int main()
{
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Initialization.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Set LED_PIN and LFO_PIN as outputs, and the rest as inputs.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    DATA_DIR_REG = 0x00;
    DATA_DIR_REG |= (1 << LED_PIN) | (1 << LFO_PIN);
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Enable pull-up resistors on input pins and drive output pins high.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    DATA_REG = 0xff;
    
    SetLfoBaseFrequency(LFO_DEFAULT_RATE);
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Set up Timer0 in fast PWM mode with no prescaler and a non-inverted compare. 
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    TCCR0A = (1 << COM0A1);                     // Clear on compare match.
    TCCR0A |= (1 << WGM01) | (1 << WGM00);      // Fast-PWM.
    TCCR0B = (1 << CS00);                       // No prescaler.
    TIMSK |= (1 << TOIE0);                      // Timer0 interrupt flagged as enabled.
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Set up Timer1 to trigger an interrupt every 1ms. CTC mode and a prescaler of 32.
    //
    // Note: In compare mode the compares are done against the OCR1C register.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    TCCR1 = (1 << CTC1);                                                    // Clear Timer on Compare match
    TCCR1 |= (1 << CS11) | (1 << CS12);                                     // CK/32
    TCNT1 = 0x00;                                                           // Starting at 0.
    OCR1C = ((CLOCK_FREQUENCY / TIMER1_PRE_SCALER) * TIMER1_FREQUENCY) - 1; // Trigger value.
    TIMSK |= (1 << OCIE1A);                                                 // Timer1 interrupt flagged as enabled
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Set up ADC input pins for waveform- and multiplier selection. VCC as reference voltage, and
    // left-aligned to get an 8-bit resolution. We're using a prescaler of 128 for an interrupt
    // frequency of 62.5kHz, and free running mode to get continuous samplings.
    // We're going to use PB3 and PB4 as inputs, meaning we need to switch between MUX[1:0] = 11
    // (PB3) and MUX[1:0] = 10 (PB4).
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    ADMUX = (1 << ADLAR) | (1 << MUX1) | (1 << MUX0);                   // Left-adjust, input channel = PB3.
    ADCSRA = (1 << ADATE) | (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0); // Auto-trigger, division = 64.
    ADCSRB = 0x00;                                                      // Free-running mode.
    
    ADCSRA |= (1 << ADEN) | (1 << ADIE);                                // ADC enabled, ADC interrupt enabled.
    ADCSRA |= (1 << ADSC);                                              // ADC start conversion.
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Set up PCINT5 (PB5) for pin change interrupts.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    PCMSK = (1 << PCINT5);                  // Select PCINT5 (PB5) a pin change interrupt source.
    GIMSK |= (1 << PCIE);                   // Enable pin change interrupts.
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Enable global interrupts. No interrupts will happen without this, regardless of individual
    // flags set elsewhere.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    sei();
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Main loop.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    while (1)
    {
        ///////////////////////////////////////////////////////////////////////////////////////////
        // Nothing here. Everything happens within the interrupt loops.
        ///////////////////////////////////////////////////////////////////////////////////////////
    }
}

/*====== G e t S w i t c h S t a t e ( ) ==============================================================================

    Purpose     : 
    Arguments   : 
    Returns     : 
    Author      : Harald Sabro
    
=====================================================================================================================*/

bool_t GetSwitchState()
{
    bool_t switch_state = SWITCH_STATE_OPEN;
    
    ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
    {
        if (INPUT_PINS & (1 << SWITCH_PIN))
        {
            switch_state = SWITCH_STATE_CLOSED;
        }
    }
    
    return switch_state;
}

/*====== D e b o u n c e S w i t c h ( ) ==============================================================================

    Purpose     : Service routine called every DEBOUNCE_CHECK_FREQUENCY milliseconds to debounce both edges.
    Arguments   : 
    Returns     : 
    Author      : Harald Sabro
    
    Note        : The code for this function was adapted from http://www.ganssle.com/debouncing-pt2.htm.
    
=====================================================================================================================*/

void DebounceSwitch(bool_t* pswitch_change,
                    bool_t* pswitch_state)
{
    static uint8_t count = DEBOUNCE_RELEASE_TIME / TIMER1_SAMPLE_RATE;
    bool_t current_switch_state;
    bool_t reset_switch_count = false;
    
    *pswitch_change = false;
    *pswitch_state = g_debounced_switch_state;
    
    current_switch_state = GetSwitchState();
    if (current_switch_state == g_debounced_switch_state)
    {
        reset_switch_count = true;
    }
    else
    {
        ///////////////////////////////////////////////////////////////////////////////////////////
        // Switch has changed. Wait for new state to become stable.
        ///////////////////////////////////////////////////////////////////////////////////////////
        
        count--;
        if (count == 0)
        {
            ///////////////////////////////////////////////////////////////////////////////////////
            // Timer expired. Accept the change and reset the timer.
            ///////////////////////////////////////////////////////////////////////////////////////
            
            g_debounced_switch_state = current_switch_state;
            *pswitch_change = true;
            *pswitch_state = current_switch_state;
            
            reset_switch_count = true;
        }
    }
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // If applicable, reset the timer.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    if (reset_switch_count == true)
    {
        if (g_debounced_switch_state == SWITCH_STATE_CLOSED)
        {
            count = DEBOUNCE_RELEASE_TIME / TIMER1_SAMPLE_RATE;
        }
        else
        {
            count = DEBOUNCE_PRESS_TIME / TIMER1_SAMPLE_RATE;
        }
    }
}

/*====== S e t L f o B a s e F r e q u e n c y ( ) ====================================================================

    Purpose     : 
    Arguments   : 
    Returns     : 
    Author      : Harald Sabro
    
=====================================================================================================================*/

void SetLfoBaseFrequency(uint16_t milliseconds)
{
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Do some boundary checking on the requested frequency. We're only going to accept frequencies
    // in the range of 0.1Hz - 20Hz.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    if ((milliseconds > LFO_MIN_TAP_TEMPO_RATE) || (milliseconds < LFO_MAX_TAP_TEMPO_RATE))
    {
        return;
    }
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Potentially save some calculating when nothing has changed (would be typical when running
    // off an external clock pulse).
    //
    // Note: May have to give or take a few milliseconds for this to really be of any value, but
    //       doesn't really hurt either.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    if (g_base_tempo != milliseconds)
    {
        ///////////////////////////////////////////////////////////////////////////////////////////
        // Store the current tap-tempo (without the multiplier). This is used both for later
        // recalculations of multiplier tempo, and driving the LED/clock pulse output.
        ///////////////////////////////////////////////////////////////////////////////////////////
        
        g_base_tempo = milliseconds;
        
        ///////////////////////////////////////////////////////////////////////////////////////////
        // Convert the given tempo to a frequency and add any multiplier.
        ///////////////////////////////////////////////////////////////////////////////////////////
        
        RecalculateLfoFrequency();
    }        
}

/*====== R e c a l c u l a t e L f o F r e q u e n c y ( ) ============================================================

    Purpose     : 
    Arguments   : 
    Returns     : 
    Author      : Harald Sabro
    
=====================================================================================================================*/

void RecalculateLfoFrequency()
{
    double new_frequency;
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Convert the given tempo from a millisecond amount to a frequency.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    new_frequency = (1000.0 / (double)g_base_tempo);
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Recalculate the base duty cycle based on the new frequency, then apply the given multiplier
    // and set the LFO duty cycle based on this.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    //
    //
    // TODO: Simplify this a bit so we don't have to always do two sets of almost the same
    //       calculations.
    //
    //
    
    g_base_duty_cycle = new_frequency * 0xffffffff / TIMER0_SAMPLE_RATE;
    
    new_frequency = new_frequency * (g_multiplier / 4.0);
    
    g_duty_cycle = new_frequency * 0xffffffff / TIMER0_SAMPLE_RATE;
}

/*====== R e s e t B a s e T e m p o ( ) ==============================================================================

    Purpose     : Reset the table counter for the LFO base tempo.
    Arguments   : 
    Returns     : 
    Author      : Harald Sabro
    
=====================================================================================================================*/

void ResetBaseTempo()
{
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Reset phase accumulator and wave table index for the base tempo.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    g_base_table_index = 0;
    g_base_phase_accumulator = 0;
}

/*====== A l i g n W a v e f o r m ( ) ================================================================================

    Purpose     : Align the LFO wave shape to a certain point based on multiplier etc.
    Arguments   : 
    Returns     : 
    Author      : Harald Sabro
    
=====================================================================================================================*/

void AlignWaveform()
{
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Align the phase accumulator appropriately based on the waveform multiplier. If we have a
    // clock we want to sync against that's running at a faster speed than the waveform we're
    // plotting, we can't sync the waveform at each clock pulse. We would never draw more than a
    // fraction of the waveform before returning resetting it.
    // Instead, when we're dealing with a slow waveform we subdivide it giving us correctly spaced
    // points along the waveform we can sync against.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Before proceeding, make sure the index isn't out of bounds.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    if (g_waveform_offset_index >= g_waveform_offset_count)
    {
        g_waveform_offset_index = 0;
    }
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Pick an offset value from the appropriate table based on the multiplier/offset count and the
    // current offset index.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    switch (g_waveform_offset_count)
    {
        case 1:     // Quart (same as clock speed) and faster.
            g_phase_accumulator = 0;
            break;
            
        case 2:     // 1/2 clock speed.
            g_phase_accumulator = g_waveform_offset_table[0][g_waveform_offset_index];
            break;
            
        case 4:     // 1/4 clock speed.
            g_phase_accumulator = g_waveform_offset_table[1][g_waveform_offset_index];
            break;
            
        default:
            g_phase_accumulator = 0;
    }
    
    g_waveform_offset_index++;
}

/*====== S e t W a v e f o r m ( ) ====================================================================================

    Purpose     : 
    Arguments   : 
    Returns     : 
    Author      : Harald Sabro
    
=====================================================================================================================*/

void SetWaveform(uint8_t voltage_reading)
{
    WAVEFORM_CHOICE temp_selection;
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Divide the voltage reading into equally spaced ranges. Based on what range the reading falls
    // under we set the waveform selection accordingly.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    if (voltage_reading >= 204)
    {
        temp_selection = SquareWave;
    }
    else if ((voltage_reading < 204) && (voltage_reading >= 153))
    {
        temp_selection = TriangleWave;
    }
    else if ((voltage_reading < 153) && (voltage_reading >= 102))
    {
        temp_selection = RampDownWave;
    }
    else if ((voltage_reading < 102) && (voltage_reading >= 51))
    {
        temp_selection = RampUpWave;
    }
    else
    {
        temp_selection = SineWave;
    }
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // We only need to update the waveform choice if it's different than what's already set.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    if (temp_selection != g_waveform)
    {
        g_waveform = temp_selection;
    }
}

/*====== S e t M u l t i p l i e r ( ) ================================================================================

    Purpose     : 
    Arguments   : 
    Returns     : 
    Author      : Harald Sabro
    
=====================================================================================================================*/

void SetMultiplier(uint8_t voltage_reading)
{
    TEMPO_MULTIPLIER temp_selection;
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Divide the voltage reading into equally spaced ranges. Based on what range the reading falls
    // under we set the multiplier selection accordingly.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    if (voltage_reading >= 204)
    {
        temp_selection = Whole;
    }
    else if ((voltage_reading < 204) && (voltage_reading >= 153))
    {
        temp_selection = Half;
    }
    else if ((voltage_reading < 153) && (voltage_reading >= 102))
    {
        temp_selection = Quarter;
    }
    else if ((voltage_reading < 102) && (voltage_reading >= 51))
    {
        temp_selection = Eighth;
    }
    else
    {
        temp_selection = Sixteenth;
    }
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // We only change the tempo multiplier if it's different than what we've already got.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    if (temp_selection != g_multiplier)
    {
        ///////////////////////////////////////////////////////////////////////////////////////////
        // In addition to setting the multiplier we also update the waveform offset count to
        // reflect how many base clock pulses goes in one waveform cycle. See ResetWaveform() for
        // more details.
        ///////////////////////////////////////////////////////////////////////////////////////////
        
        g_multiplier = temp_selection;
        
        g_waveform_offset_count = 4 / g_multiplier;
        if (g_waveform_offset_count == 0)
        {
            g_waveform_offset_count = 1;
        }
        
        RecalculateLfoFrequency();
    }
}

/*====== P l o t W a v e f o r m ( ) ==================================================================================

    Purpose     : 
    Arguments   : 
    Returns     : 
    Author      : Harald Sabro
    
=====================================================================================================================*/

void PlotWaveform()
{
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Calculate the next waveform table index. The phase accumulator keeps track of the current
    // offset within a single waveform, and the duty cycle is a fixed "step" that gets added each
    // interrupt. The given frequency will have an impact on the size of the duty cycle, and we'll
    // complete the full waveform slower or faster based on each of these steps.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    g_phase_accumulator += g_duty_cycle;
    g_table_index = (g_phase_accumulator & 0xff000000) >> 24;
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Now plot a single point on the selected waveform.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    switch (g_waveform)
    {
        case SineWave:
        
            ///////////////////////////////////////////////////////////////////////////////////////
            // Drawing this one from a table. The given index holds the plot value.
            ///////////////////////////////////////////////////////////////////////////////////////
        
            OCR0A = g_sine_table[g_table_index];
            break;
        
        case RampUpWave:
        
            ///////////////////////////////////////////////////////////////////////////////////////
            // Easily calculated; x = i
            ///////////////////////////////////////////////////////////////////////////////////////
        
            OCR0A = g_table_index;
            break;
        
        case RampDownWave:
        
            ///////////////////////////////////////////////////////////////////////////////////////
            // Easily calculated; x = max - i
            ///////////////////////////////////////////////////////////////////////////////////////
        
            OCR0A = 0xff - g_table_index;
            break;
        
        case TriangleWave:
        
            ///////////////////////////////////////////////////////////////////////////////////////
            // Easily calculated; first half: x = 2i, second half: x = max - 2i
            ///////////////////////////////////////////////////////////////////////////////////////
        
            if (g_table_index < 0x80)
            {
                OCR0A = g_table_index * 2;
            }
            else
            {
                OCR0A = 0xff - ((g_table_index - 0x80) * 2);
            }
            break;
        
        case SquareWave:
        
            ///////////////////////////////////////////////////////////////////////////////////////
            // Easily calculated; first half: x = min, second half: x = max
            ///////////////////////////////////////////////////////////////////////////////////////
        
            if (g_table_index < 0x80)
            {
                OCR0A = 0x00;
            }
            else
            {
                OCR0A = 0xff;
            }
            break;
    }
}

/*====== S t a r t T e m p o C o u n t ( ) ============================================================================

    Purpose     : 
    Arguments   : 
    Returns     : 
    Author      : Harald Sabro
    
=====================================================================================================================*/

void StartTempoCount()
{
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Sync the LFO output and start the tempo counting.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
    {
        g_is_counting_tempo = true;
        g_tempo_count = 0;
        
        AlignWaveform();
        ResetBaseTempo();
        DATA_REG &= ~(1 << LED_PIN);
    }        
}

/*====== S t o p T e m p o C o u n t ( ) ==============================================================================

    Purpose     : 
    Arguments   : 
    Returns     : 
    Author      : Harald Sabro
    
=====================================================================================================================*/

void StopTempoCount()
{
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Set the new tempo and reset the tempo counting state.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
    {
        if (g_is_counting_tempo == true)
        {
            SetLfoBaseFrequency(g_tempo_count);
            g_is_counting_tempo = false;
            g_tempo_count = 0;
        }
        
        AlignWaveform();
        ResetBaseTempo();
        DATA_REG |= (1 << LED_PIN);
    }        
}

/*====== T e m p o C o u n t T i m e o u t ( ) ========================================================================

    Purpose     : 
    Arguments   : 
    Returns     : 
    Author      : Harald Sabro
    
=====================================================================================================================*/

void TempoCountTimeout()
{
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Exit the tempo counting state without making any changes, discarding any tempo count.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
    {
        g_is_counting_tempo = false;
        g_tempo_count = 0;
    }
}

/*====== I S R ( T i m e r 0 ) ( ) ====================================================================================

    Purpose     : Timer0 overflow interrupt handler. This is where the LFO signal is generated.
    Arguments   : 
    Returns     : 
    Author      : Harald Sabro
    
    Frequency   : 31.25kHz
    
=====================================================================================================================*/

ISR(TIMER0_OVF_vect)
{
    uint8_t previous_base_table_index = g_base_table_index;
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Increase the phase accumulator by a given amount based on the required output signal
    // frequency. Then use the high 8 bits (0-255) of the phase accumulator to identify what part
    // of the wave to plot.
    //
    // We need both the base tempo (for the LED and control signal), and the actual LFO signal with
    // applied tempo multiplier.
    //
    // Also see PlotWaveform().
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    //
    //
    // TODO: Might be a shortcut just dividing by the given multiplier rather than keeping a second
    //       set of variables.
    //
    //
    
    g_base_phase_accumulator += g_base_duty_cycle;
    g_base_table_index = (g_base_phase_accumulator & 0xff000000) >> 24;
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Flag whenever there's an overflow in the base table index, i.e. the base tempo, has just
    // completed a full cycle.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    if (previous_base_table_index > g_base_table_index)
    {
        ///////////////////////////////////////////////////////////////////////////////////////////
        // As long as we're not currently in tempo counting mode (where the LED will be explicitly
        // set) toggle the LED state.
        //
        // The state of the LED output also doubles as a base frequency clock output pulse (low one
        // complete base LFO cycle, then high one complete base LFO cycle) that can be used to
        // synchronize a second (or several other) controller(s).
        ///////////////////////////////////////////////////////////////////////////////////////////
        
        if (g_is_counting_tempo == false)
        {
            DATA_REG ^= (1 << LED_PIN);
        }
    }
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Draw the next point on the waveform.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    PlotWaveform();
}

/*====== I S R ( T i m e r 1 ) ( ) ====================================================================================

    Purpose     : Timer1 compare interrupt handler. Triggers every TIMER1_FREQUENCY seconds.
    Arguments   : 
    Returns     : 
    Author      : Harald Sabro
    
    Frequency   : 1kHz
    
    Note        : Some code borrowed from http://www.ganssle.com/debouncing-pt2.htm.
    
=====================================================================================================================*/

ISR(TIMER1_COMPA_vect)
{
    bool_t switch_change = false;
    bool_t switch_state = false;
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Start by counting tempo, if applicable.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
    {
        if (g_is_counting_tempo == true)
        {
            g_tempo_count++;
        }
    }
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Make sure we don't exceed the maximum tempo length / minimum LFO frequency.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    if (g_tempo_count > LFO_MIN_TAP_TEMPO_RATE)
    {
        TempoCountTimeout();
    }
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Run each switch input through a debounce routine to make sure we get rid of any noise due to
    // the switch contacts bouncing.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    DebounceSwitch(&switch_change, &switch_state);
    if ((switch_change == true) && (switch_state == false))
    {
        if (g_is_counting_tempo == false)
        {
            ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
            {
                ///////////////////////////////////////////////////////////////////////////////////
                // Always reset the waveform on a manual tap.
                ///////////////////////////////////////////////////////////////////////////////////
                
                g_waveform_offset_index = 0;
            }
            StartTempoCount();
        }
        else
        {
            StopTempoCount();
        }
    }
}

/*====== I S R ( A D C ) ( ) ==========================================================================================

    Purpose     : Analog to Digital Converter interrupt handler.
    Arguments   : 
    Returns     : 
    Author      : Harald Sabro
    
    Frequency   : 62.5kHz (15.625kHz effective)
    
=====================================================================================================================*/

ISR(ADC_vect)
{
    static bool_t first_sampling = false;
    uint8_t sample_value;

    sample_value = ADCH;
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Because we're switching between two ADC channels in free running mode the first sampling is
    // unreliable (it may belong to either channel depending on exactly when the previous channel
    // switch was done) and we discard it in favor of the next one.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // We're running at 8MHz clock frequency with a prescaler of 128, i.e. 8MHz / 128 = 62.5kHz.
    // But we also take turns sampling two different channels, and furthermore we discard the first
    // sampling each time, meaning the actual sample frequency is 62.5kHz / 4 = 15.625kHz. Still
    // overkill considering we're only reading a static voltage...
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    if (first_sampling == true)
    {
        ///////////////////////////////////////////////////////////////////////////////////////////
        // Discard sampling.
        ///////////////////////////////////////////////////////////////////////////////////////////
        
        first_sampling = false;
    }
    else
    {
        if (ADMUX & 0x01)   // MUX[0:0] == 1.
        {
            ///////////////////////////////////////////////////////////////////////////////////////
            // Sample was read from PB3; the waveform pin.
            ///////////////////////////////////////////////////////////////////////////////////////
            
            SetWaveform(sample_value);
        }
        else    // MUX[0:0] == 0.
        {
            ///////////////////////////////////////////////////////////////////////////////////////
            // Sample was read from PB4; the multiplier pin.
            ///////////////////////////////////////////////////////////////////////////////////////
            
            SetMultiplier(sample_value);
        }
        
        ///////////////////////////////////////////////////////////////////////////////////////////
        // Switch the ADC input channel.
        ///////////////////////////////////////////////////////////////////////////////////////////
        
        ADMUX ^= (1 << MUX0);   // Toggle between MUX[1:0] = 11 and MUX[1:0] = 10.
        first_sampling = true;
    }
}

/*====== I S R ( P C I N T 5 ) ( ) ====================================================================================

    Purpose     : Pin change interrupt handler. Reads input clock / sync pulses.
    Arguments   : 
    Returns     : 
    Author      : Harald Sabro
    
=====================================================================================================================*/

ISR(PCINT0_vect)
{
    bool_t pulse_start = true;
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Detect whether this is a falling or rising edge.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    ATOMIC_BLOCK(ATOMIC_RESTORESTATE)
    {
        if (INPUT_PINS & (1 << SYNC_PIN))
        {
            pulse_start = false;
        }
    }
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    // Start or stop the tempo counting accordingly.
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    if (pulse_start == true)
    {
        StartTempoCount();
    }
    else
    {
        StopTempoCount();
    }
}

/*====== End of file ================================================================================================*/